import base64
from enum import StrEnum
from typing import Literal, NotRequired, TypedDict

from pydantic import BaseModel


class AIModel(StrEnum):
    # vertex (production Google, requires GCP environment)
    VERTEX_PRO = "vertex_ai/gemini-1.5-pro"
    VERTEX_FLASH = "vertex_ai/gemini-1.5-flash"
    VERTEX_FLASH_8B = "vertex_ai/gemini-1.5-flash-8b"
    VERTEX_FLASH_2_0 = "vertex_ai/gemini-2.0-flash"

    # gemini (publicly available, no GCP environment required)
    GEMINI_PRO = "gemini/gemini-1.5-pro"
    GEMINI_FLASH = "gemini/gemini-1.5-flash"
    GEMINI_FLASH_8B = "gemini/gemini-1.5-flash-8b"
    GEMINI_FLASH_2_0 = "gemini/gemini-2.0-flash"


class ContentType(StrEnum):
    # text
    TEXT = "text/plain"

    # images
    JPEG = "image/jpeg"
    PNG = "image/png"
    WEBP = "image/webp"

    # audio
    MP3 = "audio/mp3"
    WAV = "audio/wav"

    # files
    PDF = "application/pdf"


class LLMTextMessagePart(TypedDict):
    type: Literal["text"]
    text: str


class LLMMediaMessagePart(TypedDict):
    type: Literal["image_url"]
    image_url: str


class CacheControl(TypedDict):
    type: Literal["ephemeral"]


class LLMMessage(TypedDict):
    role: Literal["user", "assistant", "system"]
    content: list[LLMTextMessagePart | LLMMediaMessagePart]
    cache_control: NotRequired[CacheControl]


class MessagePart(BaseModel):
    content: str
    content_type: ContentType = ContentType.TEXT

    @classmethod
    def from_media(cls, *, content: bytes, content_type: ContentType) -> "MessagePart":
        return cls(content=base64.b64encode(content).decode(), content_type=content_type)

    def render(self) -> LLMTextMessagePart | LLMMediaMessagePart:
        if self.content_type == ContentType.TEXT:
            return {"type": "text", "text": self.content}
        else:
            return {"type": "image_url", "image_url": f"data:{self.content_type};base64,{self.content}"}


class UserMessage(BaseModel):
    parts: list[MessagePart]

    def render(self) -> LLMMessage:
        content: LLMMessage = {
            "role": "user",
            "content": [part.render() for part in self.parts],
        }
        if any(part.content_type != ContentType.TEXT for part in self.parts):
            content["cache_control"] = {"type": "ephemeral"}
        return content


class AssistantMessage(BaseModel):
    text: str

    def render(self) -> LLMMessage:
        return {"role": "assistant", "content": [{"type": "text", "text": self.text}]}


class SystemMessage(BaseModel):
    parts: list[MessagePart]

    def render(self) -> LLMMessage:
        return {
            "role": "system",
            "content": [part.render() for part in self.parts],
        }



from ._internal.agent import AgentResponse, IAgentLogger
from ._internal.types.agent import (
    AIModel,
    AssistantMessage,
    ContentType,
    LLMMediaMessagePart,
    LLMMessage,
    LLMTextMessagePart,
    MessagePart,
    SystemMessage,
    UserMessage,
)

__all__ = [
    "AgentResponse",
    "AIModel",
    "IAgentLogger",
    "AssistantMessage",
    "LLMMediaMessagePart",
    "LLMMessage",
    "LLMTextMessagePart",
    "SystemMessage",
    "MessagePart",
    "ContentType",
    "UserMessage",
]



import hashlib
from collections.abc import Awaitable, Callable
from typing import Any, overload

from pydantic import BaseModel

from ..errors import Honk
from .agent import Agent, AIModel, SystemMessage, UserMessage
from .conversation import Conversation
from .result import Result, TextResult
from .state import FlowRun, NodeState, get_current_flow_run
from .types.agent import AssistantMessage


class Task[**P, R: Result]:
    def __init__(
        self,
        generator: Callable[P, Awaitable[R]],
        /,
        *,
        retries: int = 0,
        adapter_model: AIModel = AIModel.GEMINI_FLASH,
    ) -> None:
        self._generator = generator
        self._retries = retries
        self._adapter_model = adapter_model
        self._adapter_model = adapter_model

    @property
    def result_type(self) -> type[R]:
        result_type = self._generator.__annotations__.get("return")
        if result_type is None:
            raise Honk(f"Task {self.name} has no return type annotation")
        return result_type

    @property
    def name(self) -> str:
        return self._generator.__name__

    async def generate(self, state: NodeState[R], *args: P.args, **kwargs: P.kwargs) -> R:
        state_hash = self.__hash_task_call(*args, **kwargs)
        if state_hash != state.last_hash:
            result = await self._generator(*args, **kwargs)
            state.add_result(result=result, new_hash=state_hash, overwrite=True)
            return result
        else:
            return state.result

    async def refine(
        self,
        *,
        user_message: UserMessage,
        context: SystemMessage | None = None,
        index: int = 0,
    ) -> R:
        flow_run = self.__get_current_flow_run()
        node_state = flow_run.get(task=self, index=index)

        if len(node_state.conversation.result_messages) == 0:
            raise Honk("Cannot refine a task that has not been initially generated")

        if context is not None:
            node_state.set_context(context=context)
        node_state.add_user_message(message=user_message)

        result = await self.__adapt(conversation=node_state.conversation, agent=flow_run.agent)
        node_state.add_result(result=result)
        flow_run.upsert_node_state(node_state)

        return result

    def edit(self, *, result: R, index: int = 0) -> None:
        flow_run = self.__get_current_flow_run()
        node_state = flow_run.get(task=self, index=index)
        node_state.edit_last_result(result=result)
        flow_run.upsert_node_state(node_state)

    def undo(self, *, index: int = 0) -> None:
        flow_run = self.__get_current_flow_run()
        node_state = flow_run.get(task=self, index=index)
        node_state.undo()
        flow_run.upsert_node_state(node_state)

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        flow_run = self.__get_current_flow_run()
        node_state = flow_run.get_next(task=self)
        result = await self.generate(node_state, *args, **kwargs)
        flow_run.upsert_node_state(node_state)
        return result

    async def __adapt(self, *, conversation: Conversation[R], agent: Agent) -> R:
        messages: list[UserMessage | AssistantMessage] = []
        for message_index in range(len(conversation.user_messages)):
            user_message = conversation.user_messages[message_index]
            result = conversation.result_messages[message_index]

            if isinstance(result, TextResult):
                assistant_text = result.text
            else:
                assistant_text = result.model_dump_json()
            assistant_message = AssistantMessage(text=assistant_text)
            messages.append(assistant_message)
            messages.append(user_message)

        return await agent(
            messages=messages,
            model=self._adapter_model,
            task_name=f"adapt--{self.name}",
            system=conversation.context,
            response_model=self.result_type,
        )

    def __hash_task_call(self, *args: P.args, **kwargs: P.kwargs) -> int:
        def update_hash(argument: Any, current_hash: Any = hashlib.sha256()) -> None:
            try:
                if isinstance(argument, list | tuple | set):
                    for item in argument:
                        update_hash(item, current_hash)
                elif isinstance(argument, dict):
                    for key, value in argument.items():
                        update_hash(key, current_hash)
                        update_hash(value, current_hash)
                elif isinstance(argument, BaseModel):
                    update_hash(argument.model_dump_json())
                elif isinstance(argument, bytes):
                    current_hash.update(argument)
                elif isinstance(argument, Agent):
                    current_hash.update(b"AGENT")
                else:
                    current_hash.update(str(argument).encode())
            except TypeError:
                raise Honk(f"Unhashable argument to task {self.name}: {argument}")

        result = hashlib.sha256()
        update_hash(args, result)
        update_hash(kwargs, result)

        return int(result.hexdigest(), 16)

    def __get_current_flow_run(self) -> FlowRun[Any]:
        run = get_current_flow_run()
        if run is None:
            raise Honk("No current flow run")
        return run


@overload
def task[**P, R: Result](generator: Callable[P, Awaitable[R]], /) -> Task[P, R]: ...
@overload
def task[**P, R: Result](
    *, retries: int = 0, adapter_model: AIModel = AIModel.GEMINI_FLASH
) -> Callable[[Callable[P, Awaitable[R]]], Task[P, R]]: ...
def task[**P, R: Result](
    generator: Callable[P, Awaitable[R]] | None = None,
    /,
    *,
    retries: int = 0,
    adapter_model: AIModel = AIModel.GEMINI_FLASH,
) -> Task[P, R] | Callable[[Callable[P, Awaitable[R]]], Task[P, R]]:
    if generator is None:

        def decorator(fn: Callable[P, Awaitable[R]]) -> Task[P, R]:
            return Task(fn, retries=retries, adapter_model=adapter_model)

        return decorator

    return Task(generator, retries=retries, adapter_model=adapter_model)



from ._internal.agent import Agent
from ._internal.flow import FlowArguments, flow
from ._internal.result import Result, TextResult
from ._internal.task import task

__all__ = ["Agent", "flow", "FlowArguments", "Result", "TextResult", "task"]
